from apispec import APISpec
from apispec.ext.marshmallow import MarshmallowPlugin
from flask import Flask, session
from flask_apispec import FlaskApiSpec
from flask_restful import Api

from openfabric_pysdk.config import manifest, state_config
from openfabric_pysdk.toolset import ConfigRestApi, ManifestRestApi, BenchmarkRestApi
from openfabric_pysdk.transport.queue import *
from openfabric_pysdk.transport.rest import ExecutionRestApi
from openfabric_pysdk.transport.socket import Socket


#######################################################
#  Application
#######################################################
class Application:
    state: State = None
    __api: Api = None
    __app: Flask = None
    __socket: Socket = None
    __docs: FlaskApiSpec = None

    # ------------------------------------------------------------------------
    def __init__(self, app: Flask):
        self.__app = app
        self.__api = Api(app)
        self.__docs = FlaskApiSpec(app)
        self.state = State()

    # ------------------------------------------------------------------------
    def install_specs__rest(self, endpoint):
        print(f"Install Specs REST endpoints on {endpoint}")
        specs = {
            'APISPEC_SPEC': APISpec(
                title="App " + manifest.get('name'),
                version=manifest.get('version'),
                plugins=[MarshmallowPlugin()],
                openapi_version='2.0.0',
                info=dict(
                    termsOfService='https://openfabric.ai/terms/',
                    contact=dict(
                        name=manifest.get('organization'),
                        url="https://openfabric.ai"
                    ),
                    description=manifest.get('description')),
            ),
            'APISPEC_SWAGGER_URL': f'/{endpoint}/',  # URI to access API Doc JSON
            'APISPEC_SWAGGER_UI_URL': f'/{endpoint}-ui/'  # URI to access UI of API Doc
        }
        self.__app.config.update(specs)

    # ------------------------------------------------------------------------
    def install_execution_rest(self, endpoint):
        print(f"Install Execution REST endpoints on {endpoint}")

        resource_args = dict(state=self.state)

        self.__api.add_resource(ExecutionRestApi, endpoint, resource_class_kwargs=resource_args)
        self.__docs.register(ExecutionRestApi, resource_class_kwargs=resource_args)

    # ------------------------------------------------------------------------
    def install_config_rest(self, endpoint):
        if ConfigSchema is None:
            print(f"No Config schema available")
            return
        print(f"Install Config REST endpoints on {endpoint}")

        resource_args = dict(state=self.state)

        self.__api.add_resource(ConfigRestApi, endpoint, resource_class_kwargs=resource_args)
        self.__docs.register(ConfigRestApi, resource_class_kwargs=resource_args)

    # ------------------------------------------------------------------------
    def install_manifest_rest(self, endpoint):
        print(f"Install Manifest REST endpoints on {endpoint}")
        self.__api.add_resource(ManifestRestApi, endpoint)
        self.__docs.register(ManifestRestApi)

    # ------------------------------------------------------------------------
    def install_benchmark_rest(self, endpoint):
        print(f"Install Benchmark REST endpoints on {endpoint}")
        self.__api.add_resource(BenchmarkRestApi, endpoint)
        self.__docs.register(BenchmarkRestApi)

    # ------------------------------------------------------------------------
    def install_execution_queue(self, endpoint):
        print(f"Install Execution REST Queue endpoints on {endpoint}")

        resource_args = dict(state=self.state)

        self.__api.add_resource(QueueApi_Get, endpoint + "/get", resource_class_kwargs=resource_args)
        self.__docs.register(QueueApi_Get, resource_class_kwargs=resource_args)

        self.__api.add_resource(QueueApi_List, endpoint + "/list", resource_class_kwargs=resource_args)
        self.__docs.register(QueueApi_List, resource_class_kwargs=resource_args)

        self.__api.add_resource(QueueApi_Post, endpoint + "/post", resource_class_kwargs=resource_args)
        self.__docs.register(QueueApi_Post, resource_class_kwargs=resource_args)

        self.__api.add_resource(QueueApi_Delete, endpoint + "/delete", resource_class_kwargs=resource_args)
        self.__docs.register(QueueApi_Delete, resource_class_kwargs=resource_args)

    # ------------------------------------------------------------------------
    def install_execution_socket(self, endpoint):
        print(f"Install Execution SOCKET endpoints on {endpoint}")
        self.__socket = Socket(endpoint, session, self.__app, self.state)

    # ------------------------------------------------------------------------
    def install_configuration(self):
        print(f"Install APP configuration")
        if config_callback_function:
            items = state_config.all().items()
            config = dict(map(lambda kv: (kv[0], ConfigSchema().load(kv[1])), items))
            if config:
                config_callback_function(config, self.state)

    # ------------------------------------------------------------------------
    def run(self, debug, host, port):
        self.__socket.run(debug=debug, host=host, port=port)
        # self.__app.run(debug=debug, host=host, port=port)



import json
import logging
import zlib
from hashlib import md5

from typing import Any, Set, Dict

import eventlet
from flask import Flask, request
from flask_socketio import SocketIO, Namespace, emit

from openfabric_pysdk.context import RaySchema, Ray, State, StateSchema
from openfabric_pysdk.engine import foreground, background
from openfabric_pysdk.loader import *
from openfabric_pysdk.utility import ChangeDetector


#######################################################
#  Execution Socket
#######################################################
class Socket(Namespace):
    __app = None
    __state = None
    __session = None
    __socket_io = None
    __active_sessions: Set[str] = None

    # ------------------------------------------------------------------------
    def __init__(self, socket_namespace, socket_session, app: Flask, state: State):
        super().__init__(socket_namespace)
        self.__session = socket_session
        # Set this variable to "threading", "eventlet" or "gevent" to test the
        # different async modes, or leave it set to None for the application to choose
        # the best option based on installed packages.
        async_mode = "eventlet"
        self.__socket_io = SocketIO(app, async_mode=async_mode, cors_allowed_origins='*')
        self.__socket_io.on_namespace(self)
        self.__app = app
        self.__state = state
        self.__active_sessions = set()

    # ------------------------------------------------------------------------
    def run(self, debug, host, port):
        self.__socket_io.run(self.__app, host=host, debug=debug, port=port)

    # ------------------------------------------------------------------------
    def on_execute(self, data: Any, background: bool):
        with MeasureBlockTime("Socket::execute"):
            # Uncompress data
            with MeasureBlockTime("Socket::decompress"):
                uncompressed = zlib.decompress(data)
                request_payload: str = uncompressed.decode('utf-8')
                dictionary: Dict[str, Any] = json.loads(request_payload)

            sid = request.sid
            data = dictionary.get('body', None)
            header = dictionary.get('header', dict())
            uid = header.get('uid', None)

            if background is True:
                self.__execute_background(data, sid, uid)
            else:
                self.__execute_foreground(data, sid, uid)

    def __execute_foreground(self, data: str, sid: str, uid: str):
        # Setup
        qid = foreground.prepare(data, qid=sid, sid=sid, uid=uid, state=self.__state)
        ray = foreground.ray(qid)
        # Execute in foreground
        with MeasureBlockTime("Socket::callback"):
            foreground.process(qid)
            output = foreground.read(qid, 'out')
            ray_dump = RaySchema().dump(ray)
            foreground.delete(qid)
            emit('response', dict(output=output, ray=ray_dump))

    def __execute_background(self, data: str, sid: str, uid: str):
        # Setup
        qid = background.prepare(data, sid=sid, uid=uid, state=self.__state)
        ray = background.ray(qid)
        emit('submitted', RaySchema().dump(ray))
        # Execute in background
        with MeasureBlockTime("Socket::callback"):
            while sid in self.__active_sessions:
                ray_dump = RaySchema().dump(ray)
                emit('progress', ray_dump)
                if ray.finished is True:
                    # input = background.get(qid, 'in')
                    output = background.read(qid, 'out')
                    emit('response', dict(output=output, ray=ray_dump))
                    break
                eventlet.sleep(0.1)

    # ------------------------------------------------------------------------
    def on_resume(self, uid: str):
        sid = request.sid
        with MeasureBlockTime("OpenfabricSocket::restore"):
            def criteria(_ray: Ray):
                # is deleted ?
                if _ray is None:
                    return False
                # is different user ?
                if _ray.uid != uid:
                    return False
                # is an active session available ?
                if _ray.sid in self.__active_sessions and _ray.sid != sid:
                    return False
                return True

            # Filter rays
            rays = background.rays(criteria)
            for ray in rays:
                emit('progress', RaySchema().dump(ray))

    # ------------------------------------------------------------------------
    def on_restore(self, qid: str):
        ray_dump = foreground.read(qid, 'ray')
        input_dump = foreground.read(qid, 'in')
        output_dump = foreground.read(qid, 'out')
        emit('restore', dict(input=input_dump, output=output_dump, ray=ray_dump))

    # ------------------------------------------------------------------------
    def on_state(self, uid: str):
        sid = request.sid
        with MeasureBlockTime("Socket::state"):
            while sid in self.__active_sessions:
                changed = ChangeDetector.is_changed('app::state' + sid, self.__state, StateSchema().dump)
                if changed is True:
                    emit('state', StateSchema().dump(self.__state))
                eventlet.sleep(0.1)

    # ------------------------------------------------------------------------
    def on_delete(self, qid: str):
        sid = request.sid
        with MeasureBlockTime("OpenfabricSocket::delete"):
            foreground.delete(qid)
            background.delete(qid)

    # ------------------------------------------------------------------------
    def on_connect(self):
        sid = request.sid
        logging.info(f'Client connected {sid} on {request.host}')
        self.__active_sessions.add(sid)

    # ------------------------------------------------------------------------
    def on_disconnect(self):
        sid = request.sid
        logging.info(f'Client disconnected {sid} on {request.host}')
        self.__active_sessions.remove(sid)



import collections
from hashlib import md5
from typing import OrderedDict, Any, Dict


#######################################################
#  LRU
#######################################################
class LRU:
    __size: int = None
    __evict = None
    __lru_cache: OrderedDict[str, Any] = None

    # ------------------------------------------------------------------------
    def __init__(self, size, evict=None):
        self.__size = size
        self.__lru_cache = collections.OrderedDict()
        self.__evict = evict

    # ------------------------------------------------------------------------
    def get(self, key: str, default=None):
        try:
            value = self.__lru_cache.pop(key)
            self.__lru_cache[key] = value
            return value
        except KeyError:
            return default

    # ------------------------------------------------------------------------
    def put(self, key: str, value: Any):
        try:
            self.__lru_cache.pop(key)
        except KeyError:
            if len(self.__lru_cache) >= self.__size:
                _key, _val = self.__lru_cache.popitem(last=False)
                if self.__evict is not None:
                    self.__evict(_val)
        self.__lru_cache[key] = value

    # ------------------------------------------------------------------------
    def rem(self, key: str):
        if self.__lru_cache[key] is not None:
            del self.__lru_cache[key]


#######################################################
#  Change detector
#######################################################
class ChangeDetector:
    changes = LRU(500)

    @classmethod
    def is_changed(cls, name, content, serializer=None):
        if content is not None:
            content = serializer(content) if serializer is not None else content
        else:
            content = ''
        content_hash = cls.changes.get(name)
        new_content_hash = md5(str(content).encode('utf-8')).hexdigest()
        cls.changes.put(name, new_content_hash)
        return content_hash != new_content_hash


#######################################################
#  Change detector
#######################################################
class SchemaUtil:

    @staticmethod
    def create(instance: Any, data: Dict[Any, Any]):
        for key in data:
            instance.__setattr__(key, data[key])
        return instance



import logging
import pathlib
import traceback

from flask import Flask, app, render_template
from flask_cors import CORS

from openfabric_pysdk.config import manifest
from openfabric_pysdk.application import Application
from openfabric_pysdk.context import StateStatus

FlaskApp = Flask(__name__, template_folder=f"{pathlib.Path(__file__).parent}/templates")
CORS(FlaskApp)


@FlaskApp.route("/")
def index():
    return render_template("index.html", manifest=manifest.all())


class Starter:

    @staticmethod
    def ignite(debug, host, port=5000):
        # Setup logger
        logger = logging.getLogger()
        logger.setLevel(logging.INFO)
        application = Application(FlaskApp)
        application.state.status = StateStatus.STARTING

        try:
            # Setup socket services
            application.install_execution_socket('/app')
            # Setup reset services
            application.install_execution_rest('/app')
            application.install_config_rest('/config')
            application.install_specs__rest('/swagger')
            application.install_manifest_rest('/manifest')
            application.install_benchmark_rest('/benchmark')
            application.install_execution_queue('/queue')
            # Install configuration
            application.install_configuration()
            application.state.status = StateStatus.RUNNING
            # Run
            application.run(debug=debug, host=host, port=port)
        except:
            application.state.status = StateStatus.CRASHED
            print(f"Failed starting app")
            traceback.print_exc()



import uuid

from flask_apispec import marshal_with, doc, use_kwargs
from flask_apispec.views import MethodResource
from flask_restful import Resource

from openfabric_pysdk.engine import foreground
from openfabric_pysdk.loader import *
from openfabric_pysdk.context import State


#######################################################
#  Execution API
#######################################################
class ExecutionRestApi(MethodResource, Resource):
    __state: State = None

    # ------------------------------------------------------------------------
    def __init__(self, state: State = None):
        self.__state = state

    @doc(description="Execute app and get response", tags=["Execution"])
    @use_kwargs(InputSchema, location='json')
    @marshal_with(OutputSchema)  # marshalling
    def post(self, *args) -> OutputClass:
        sid = uuid.uuid4().hex
        data = InputSchema().dump(list(args) if InputSchema().many is True else args[0])
        qid = foreground.prepare(data, sid=sid, state=self.__state)
        return foreground.process(qid)



from datetime import datetime
from enum import Enum
from typing import Dict, List, Set

from marshmallow import Schema, fields, post_load
from tqdm.asyncio import tqdm, tqdm_asyncio

from openfabric_pysdk.utility import SchemaUtil


#######################################################
#  Ray Status
#######################################################
class RayStatus(Enum):
    QUEUED = 'queued',
    PENDING = 'pending',
    COMPLETED = 'completed',
    RUNNING = 'running',
    CANCELED = 'canceled',
    REMOVED = 'removed',
    UNKNOWN = 'unknown',
    FAILED = 'failed'

    # ------------------------------------------------------------------------
    def __str__(self):
        return self.name


#######################################################
#  Baar
#######################################################
class Bar:
    percent: str = 0
    remaining: str = 'NaN'


class BarSchema(Schema):
    percent = fields.String(allow_none=True)
    remaining = fields.String(allow_none=True)

    @post_load
    def create(self, data, **kwargs):
        return SchemaUtil.create(Bar(), data)


#######################################################
#  Message
#######################################################
class MessageType(Enum):
    INFO = 'info',
    ERROR = 'error',
    WARNING = 'warning'

    # ------------------------------------------------------------------------
    def __str__(self):
        return self.name


class Message:
    type: MessageType = 0
    content: str = None
    created_at: datetime = datetime.now()

    # ------------------------------------------------------------------------
    def __eq__(self, other):
        if not isinstance(other, Message):
            return NotImplemented
        return self.type == other.type and self.content == other.content

    def __key(self):
        return self.type, self.content

    def __hash__(self):
        return hash(self.__key())


class MessageSchema(Schema):
    type = fields.Enum(MessageType)
    content = fields.String()
    created_at = fields.DateTime()

    @post_load
    def create(self, data, **kwargs):
        return SchemaUtil.create(Message(), data)


#######################################################
#  Ray
#######################################################
class RaySchema(Schema):
    sid = fields.String()
    uid = fields.String()
    qid = fields.String()
    bars = fields.Dict(
        keys=fields.Str(),
        values=fields.Nested(BarSchema),
        allow_none=True
    )
    messages = fields.Nested(MessageSchema(many=True))
    status = fields.Enum(RayStatus)
    created_at = fields.DateTime()
    updated_at = fields.DateTime()

    @post_load
    def create(self, data, **kwargs):
        return SchemaUtil.create(Ray(None), data)


class Ray:
    uid: str = None
    sid: str = None
    qid: str = None
    finished: bool = None
    bars: Dict[str, Bar] = None
    status: RayStatus = None
    created_at: datetime = None
    updated_at: datetime = None
    messages: Set[Message] = None
    __tqdms: Dict[str, tqdm_asyncio] = None

    # ------------------------------------------------------------------------
    def __init__(self, qid):
        self.__tqdms = dict()
        self.qid = qid
        self.bars = dict(default=Bar())
        self.finished = False
        self.status = RayStatus.UNKNOWN
        self.created_at = datetime.now()
        self.updated_at = datetime.now()
        self.messages = set()

    # ------------------------------------------------------------------------
    def progress(self, name='default', step=1, total=100) -> tqdm_asyncio:
        if self.__tqdms.get(name) is None:
            self.__tqdms[name] = tqdm(total=total)
        tqdm_bar = self.__tqdms[name]
        # --
        bar = self.bars.get(name, Bar())
        self.bars[name] = bar
        f_dict = tqdm_bar.format_dict
        rate = f_dict.get("rate")
        total = tqdm_bar.total
        n = tqdm_bar.n
        remaining = (total - n) / rate if rate and total else 0
        bar.remaining = max(0, remaining)
        bar.percent = n
        # --
        tqdm_bar.update(step)
        return tqdm_bar

    # ------------------------------------------------------------------------
    def message(self, message_type: MessageType, content: str):
        message = Message()
        message.type = message_type
        message.content = content
        self.messages.add(message)

    # ------------------------------------------------------------------------
    def clear_messages(self):
        self.messages.clear()

    # ------------------------------------------------------------------------
    def tqdms(self):
        return self.__tqdms


#######################################################
#  Execution State
#######################################################
class StateStatus(Enum):
    UNKNOWN = 'unknown',
    STARTING = 'starting',
    RUNNING = 'running',
    CRASHED = 'crashed',
    PENDING_CONFIG = 'pending_config'

    # ------------------------------------------------------------------------
    def __str__(self):
        return self.name


class State:
    status: StateStatus = None
    started_at: datetime = None

    # ------------------------------------------------------------------------
    def __init__(self):
        self.status = StateStatus.UNKNOWN
        self.started_at = datetime.now()


class StateSchema(Schema):
    status = fields.Enum(StateStatus)
    started_at = fields.DateTime()



import os

from flask_apispec import marshal_with, doc, use_kwargs
from flask_apispec.views import MethodResource
from flask_restful import Resource
from marshmallow import fields, Schema, post_load
from webargs.flaskparser import use_args

from openfabric_pysdk import benchmark
from openfabric_pysdk.config import manifest, state_config
from openfabric_pysdk.context import State
from openfabric_pysdk.loader import *
from openfabric_pysdk.store import KeyValueDB
from openfabric_pysdk.utility import SchemaUtil


#######################################################
#  Config API
#######################################################

class UserIdClass(Schema):
    uid: str = None


class UserIdSchema(Schema):
    uid = fields.String(required=True)

    @post_load
    def create(self, data, **kwargs):
        return SchemaUtil.create(UserIdClass(), data)


class ConfigRestApi(MethodResource, Resource):
    __state: State = None

    # ------------------------------------------------------------------------
    def __init__(self, state: State = None):
        self.__state = state

    @doc(description="Get APP configuration", tags=["Developer"])
    @use_args(UserIdSchema, location='query')
    @marshal_with(ConfigSchema)
    def get(self, uidc: UserIdClass) -> ConfigClass:
        with MeasureBlockTime("ConfigRestApi::get"):
            config = state_config.get(uidc.uid)
            if config is False:
                return dict()
            return config

    @doc(description="Set APP configuration", tags=["Developer"])
    @use_args(UserIdSchema, location='query')
    @use_args(ConfigSchema, location='json')
    @marshal_with(ConfigSchema)
    def post(self, uidc: UserIdClass, config: ConfigClass) -> ConfigClass:

        with MeasureBlockTime("ConfigRestApi::post"):
            config = list(config) if ConfigSchema().many is True else config[0]
            state_config.set(uidc.uid, ConfigSchema().dump(config))
            if config_callback_function:
                config = dict(map(lambda kv: (kv[0], ConfigSchema().load(kv[1])), state_config.all().items()))
                config_callback_function(config, self.__state)
            return config


#######################################################
#  Manifest API
#######################################################
class ManifestSchema(Schema):
    name = fields.String(description="App name")
    version = fields.String(description="App version")
    description = fields.String(description="APP description")
    organization = fields.String(description="APP organization")
    sdk = fields.String(description="APP sdk")
    overview = fields.String(description="APP overview")
    input = fields.String(description="APP input")
    output = fields.String(description="APP output")

    def __init__(self):
        super().__init__(many=False)


class ManifestRestApi(MethodResource, Resource):

    @doc(description="Get APP manifest", tags=["Developer"])
    @marshal_with(ManifestSchema)  # marshalling
    def get(self):
        with MeasureBlockTime("ManifestRestApi::get"):
            return manifest.all()


#######################################################
#  Benchmark API
#######################################################
class BenchmarkSchema(Schema):
    name = fields.String()
    avg = fields.String()
    count = fields.String()
    stddev = fields.String()
    min = fields.String()
    max = fields.String()

    def __init__(self):
        super().__init__(many=True)


class BenchmarkRestApi(MethodResource, Resource):

    @doc(description="Get APP benchmarks", tags=["Developer"])
    @marshal_with(BenchmarkSchema)  # marshalling
    def get(self):
        with MeasureBlockTime("BenchmarkRestApi::get"):
            return benchmark.get_all_timings_json()



import os
import threading
import traceback
import uuid
from time import sleep
from typing import Dict, Any, List

from openfabric_pysdk.benchmark import MeasureBlockTime
from openfabric_pysdk.context import RayStatus, Ray, RaySchema, State
from openfabric_pysdk.loader import InputSchema, execution_callback_function, OutputSchema
from openfabric_pysdk.store import Store
from openfabric_pysdk.task import Task


#######################################################
#  Foreground Engine
#######################################################
class ForegroundEngine:
    __store: Store = None
    __state: State = None
    __rays: Dict[str, Ray] = None

    # ------------------------------------------------------------------------
    def __init__(self):
        self.__rays = dict()
        self.__store = Store(path=f"{os.getcwd()}/datastore")

    # ------------------------------------------------------------------------
    def prepare(self, data: str, qid=None, sid=None, uid=None, state: State = None) -> str:
        if qid is None:
            qid: str = uuid.uuid4().hex
        ray = self.ray(qid)
        ray.status = RayStatus.QUEUED
        ray.sid = sid
        ray.uid = uid
        self.write(qid, 'ray', ray, RaySchema().dump)
        self.write(qid, 'in', data)
        self.__state = state
        return qid

    # ------------------------------------------------------------------------
    def ray(self, qid: str):
        if self.__rays.get(qid) is None:
            ray = Ray(qid=qid)
            self.__rays[qid] = ray
        return self.__rays[qid]

    # ------------------------------------------------------------------------
    def rays(self, criteria=None) -> List[Ray]:
        rays: List[Ray] = []
        for qid, ray in self.__rays.items():
            if criteria is None or criteria(ray):
                rays.append(ray)
        return rays

    # ------------------------------------------------------------------------
    def process(self, qid):

        with MeasureBlockTime("Engine::execution_callback_function"):

            output = None
            ray = self.ray(qid)
            try:
                data = self.read(qid, 'in', InputSchema().load)
                if data is None:
                    return None

                ray.status = RayStatus.RUNNING
                self.write(qid, 'ray', ray, RaySchema().dump)
                output = execution_callback_function(data, ray, self.__state)
                ray.status = RayStatus.COMPLETED
                ray.progress(step=100)
            except Exception as e:
                print(f"Failed executing: [{qid}]{e}")
                traceback.print_exc()
                ray.status = RayStatus.FAILED
        ray.finished = True
        self.write(qid, 'ray', ray, RaySchema().dump)
        self.write(qid, 'out', output, OutputSchema().dump)

        return output

    # ------------------------------------------------------------------------
    def read(self, qid: str, key: str, deserializer=None) -> Any:
        output = self.__store.get(qid, key)
        if output is None:
            return None
        output = deserializer(output) if deserializer is not None else output
        return output

    # ------------------------------------------------------------------------
    def write(self, qid: str, key: str, val: Any, serializer=None):
        if val is not None:
            val = serializer(val) if serializer is not None else val
        self.__store.set(qid, key, val)

    # ------------------------------------------------------------------------
    def delete(self, qid: str) -> Ray:
        ray = self.ray(qid)
        self.__store.drop(qid)
        ray.status = RayStatus.REMOVED
        return ray


#######################################################
#  Background Engine
#######################################################
class BackgroundEngine:
    __task: Task = None
    __instances: int = 0
    __running: bool = False
    __current_qid: str = None
    __worker: threading.Thread = None
    __lock: threading.Condition = threading.Condition()

    # ------------------------------------------------------------------------
    def __init__(self):
        self.__lock.acquire()
        if self.__instances == 0:
            self.__task = Task()
            self.__worker = threading.Thread(target=self.__process, args=())
            self.__worker.start()

        self.__instances = self.__instances + 1
        self.__lock.release()

        # Wait for processing thread to start
        while not self.__running:
            sleep(0.1)

    # ------------------------------------------------------------------------
    def __del__(self):
        self.__lock.acquire()
        if self.__instances > 0:
            self.__lock.release()
            return

        self.__running = False

        self.__lock.notify_all()
        self.__lock.release()

    # ------------------------------------------------------------------------
    def __process(self):
        self.__running = True
        while self.__running:
            self.__lock.acquire()
            self.__current_qid = None
            while self.__running and self.__task.empty():
                self.__lock.wait()
            try:
                self.__current_qid = self.__task.next()
            except:
                print("Queue empty!")
                traceback.print_exc()
            finally:
                self.__lock.release()

            if self.__running and self.__current_qid is not None:
                foreground.process(self.__current_qid)

    # ------------------------------------------------------------------------
    def prepare(self, data: str, qid=None, sid=None, uid=None, state: State = None) -> str:
        self.__lock.acquire()
        qid: str = foreground.prepare(data, qid=qid, sid=sid, uid=uid, state=state)
        self.__task.add(qid)
        self.__lock.notify_all()
        self.__lock.release()
        return qid

    # ------------------------------------------------------------------------
    def ray(self, qid: str) -> Ray:
        return foreground.ray(qid)

    # ------------------------------------------------------------------------
    def rays(self, criteria=None) -> List[str]:
        rays: List[str] = []
        for qid in self.__task.all():
            ray = foreground.read(qid, 'ray', RaySchema().load)
            if criteria is None or criteria(ray):
                rays.append(ray)
        rays.sort(key=lambda r: r.created_at)
        return rays

    # ------------------------------------------------------------------------
    def delete(self, qid: str) -> Ray:
        self.__lock.acquire()
        self.__task.rem(qid)
        ray = foreground.delete(qid)
        self.__lock.notify_all()
        self.__lock.release()
        return ray

    # ------------------------------------------------------------------------
    def read(self, qid: str, key: str, deserializer=None) -> Any:
        return foreground.read(qid, key, deserializer)

    # ------------------------------------------------------------------------
    def write(self, qid: str, key: str, val: Any, serializer=None):
        return foreground.write(qid, key, val, serializer)


foreground = ForegroundEngine()
background = BackgroundEngine()



import importlib
from pydoc import locate

from openfabric_pysdk.benchmark import MeasureBlockTime
from openfabric_pysdk.config import execution


#######################################################
# Class loader utility class
#######################################################
class Loader:

    @staticmethod
    def import_class(name):
        return locate(name.__module__ + "." + name.__name__)

    @staticmethod
    def get_class(name):
        with MeasureBlockTime("Loader::get_class"):
            class_config = execution.get(name)
            if class_config is None:
                return None
            clazz = class_config['class']
            package = class_config['package']
            if clazz is None or package is None:
                return None
            module = importlib.import_module(package)
            return locate(module.__name__ + "." + clazz)

    @staticmethod
    def get_function(name):
        with MeasureBlockTime("Loader::get_function"):
            function_config = execution.get(name)
            if function_config is None:
                return None
            function = function_config['function']
            package = function_config['package']
            if function is None or package is None:
                return None
            module = importlib.import_module(package)
            print(module.__name__ + "." + function)
            return locate(module.__name__ + "." + function)


# Output concept
InputClass = Loader.get_class("input_class")
InputSchema = Loader.get_class("input_schema")

# Output concept
OutputClass = Loader.get_class("output_class")
OutputSchema = Loader.get_class("output_schema")

# Config concept
ConfigClass = Loader.get_class("config_class")
ConfigSchema = Loader.get_class("config_schema")

# Execution callback function
execution_callback_function = Loader.get_function("main_callback")

# Config callback function
config_callback_function = Loader.get_function("config_callback")
